from cassandra.cqlengine import management as db_management

from django.db.backends.schema import (
    BaseDatabaseSchemaEditor
)

from djangocassandra.db.models import (
    get_column_family,
    internal_type_to_column_map
)


class CassandraSchemaEditor(BaseDatabaseSchemaEditor):
    known_models = set()

    def skip_default(
        self,
        field
    ):
        return True

    def _create_db_table(
        self,
        column_family
    ):
        db_management.sync_table(column_family)

    def create_model(
        self,
        model
    ):
        column_family = get_column_family(
            self.connection,
            model
        )
        self._create_db_table(column_family)

    def delete_model(
        self,
        model
    ):
        column_family = get_column_family(
            self.connection,
            model
        )

        db_management.drop_table(column_family)

    def alter_unique_together(
        self,
        model,
        old_unique_together,
        new_unique_together
    ):
        '''
        Unique together isn't a concept Cassandra understands
        out of the box. We will just do a noop in this case.
        '''
        pass

    def alter_index_together(
        self,
        model,
        old_index_together,
        new_index_together
    ):
        '''
        Index together is a similar concept to Cassandra's compound
        primary keys but to maintain compatibilty with Django projects
        that are migrating from some other database backend I will keep
        these concepts seperate instead of trying to overload
        _meta.index_together
        '''
        pass

    def alter_db_table(
        self,
        model,
        old_db_table,
        new_db_table
    ):
        '''
        Renaming the table isn't supported by Cassandra. This
        has to be done in a manual process as follows:
            
            1) create schema for NEW_Keyspace
            2) stop writes to OLD_Keyspace from app (reads can continue)
            3) flush OLD_Keyspace on every node, via nodetool
            4) hard link all sstables from OLD_Keyspace directory to
               NEW_Keyspace directory
            5) call nodetool -h localhost refresh NEW_Keyspace
            6) enable reads/writes from/to NEW_Keyspace from app
               (disable reads on OLD_Keyspace)
            7) clean up OLD_Keyspace (drop schema, delete files, etc.)
        '''
        pass

    def alter_db_tablespace(
        self,
        model,
        old_db_tablespace,
        new_db_tablespace
    ):
        '''
        Moving a model to a different keyspace requires
        creating the column family in the new keyspace
        (new_db_tablespace), migrating the data and then
        deleting the old column family.
        '''
        pass

    def add_field(
        self,
        model,
        field
    ):
        column_family = self._add_field(
            model,
            field
        )
        self._create_db_table(column_family)

    def _add_field(
        self,
        model,
        field,
        column_family=None
    ):
        if None is column_family:
            column_family = get_column_family(
                self.connection,
                model
            )

        cql_field_type = internal_type_to_column_map[field.get_internal_type()]
        cql_field = cql_field_type(
            primary_key=field.primary_key,
            index=field.db_index,
            db_field=field.name,
            required=not field.blank
        )
        setattr(column_family, field.name, cql_field)
        return column_family
        
    def remove_field(
        self,
        model,
        field
    ):
        column_family = self._remove_field(
            model,
            field
        )
        self._create_db_table(column_family)

    def _remove_field(
        self,
        model,
        field,
        column_family=None
    ):
        if None is column_family:
            column_family = get_column_family(
                self.connection,
                model
            )

        delattr(column_family, field.name)
        return column_family

    def alter_field(
        self,
        model,
        old_field,
        new_field,
        strict=False
    ):
        '''
        A little more care needs to be taken here to handle
        cases where exceptions are raised in either _remove_field
        or in _add_field. This could leave the column_family
        registery dirty and in a state that would require the
        web service to restart to function properly.
        '''
        column_family = get_column_family(
            self.connection,
            model
        )
        self._remove_field(
            model,
            old_field,
            column_family=column_family
        )
        self._add_field(
            model,
            new_field,
            column_family=column_family
        )

        self._create_db_table(column_family)



import uuid

from django.db.models import AutoField


class AutoFieldUUID(AutoField):
    def to_python(
        self,
        value
    ):
        if (
            value is None or
            isinstance(value, uuid.UUID)
        ):
            return value

        try:
            return uuid.UUID(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )

    def get_prep_value(self, value):
        value = super(AutoField, self).get_prep_value(value)
        if (
            value is None or
            isinstance(value, uuid.UUID)
        ):
            return value

        return uuid.UUID(value)



import itertools

from django.db.backends import BaseDatabaseIntrospection
from djangotoolbox.db.base import NonrelDatabaseIntrospection


class DatabaseIntrospection(NonrelDatabaseIntrospection):
    def get_table_list(
        self,
        cursor=None
    ):
        if None is cursor:
            cursor = self.connection.cursor()

        current_keyspace = cursor.keyspace

        keyspaces = [
            key for key in self.connection.settings_dict.get(
                'KEYSPACES'
            ).keys()
        ]

        table_list = []
        try:
            cursor.set_keyspace('system')
            for keyspace in keyspaces:
                table_list = itertools.chain(
                    table_list, cursor.execute(
                        ''.join([
                            'SELECT columnfamily_name from ',
                            'schema_columnfamilies where keyspace_name=\'',
                            keyspace,  # TODO: SANITIZE ME JUST IN CASE!!!
                            '\''
                        ])
                    )
                )

        finally:
            if None is not current_keyspace:
                cursor.set_keyspace(current_keyspace)

        return [row['columnfamily_name'] for row in table_list]

    def table_names(self, cursor):
        return BaseDatabaseIntrospection.table_names(self, cursor)



import itertools

from uuid import uuid4

from django.db.utils import (
    ProgrammingError,
    NotSupportedError
)

from django.db.models.sql.constants import MULTI
from django.db.models.sql.where import (
    WhereNode,
    AND,
    OR
)

from djangotoolbox.db.basecompiler import (
    NonrelQuery,
    NonrelCompiler,
    NonrelInsertCompiler,
    NonrelUpdateCompiler,
    NonrelDeleteCompiler
)

from cassandra.cqlengine.query import (
    BatchQuery,
    QueryException
)

from djangocassandra.db.models import (
    get_column_family
)

from .predicate import (
    CompoundPredicate,
    COMPOUND_OP_AND,
    COMPOUND_OP_OR
)

from .utils import (
    safe_call,
    sort_rows
)


class CassandraQuery(NonrelQuery):
    MAX_RESULT_COUNT = 10000

    def __init__(
        self,
        compiler,
        fields
    ):
        super(CassandraQuery, self).__init__(
            compiler,
            fields
        )

        self.meta = self.query.get_meta()
        
        if hasattr(self.query.model, 'CassandraMeta'):
            self.cassandra_meta = self.query.model.CassandraMeta

        else:
            self.cassandra_meta = None

        self.pk_column = self.meta.pk.column
        self.column_family = self.meta.db_table
        self.columns = fields
        self.where = None
        self.cache = None
        self.allows_inefficient = True  # TODO: Make this a config setting
        self.ordering = []
        self.filters = []
        self.inefficient_ordering = []

        self.connection.ensure_connection()
        self.column_family_class = get_column_family(
            self.connection,
            self.query.model
        )

        self.column_names = [
            field.db_column if field.db_column else field.column
            for field in fields
        ]
        self.indexed_columns = [
            field.db_column if field.db_column else field.column
            for field in fields if field.db_index
        ]

        if hasattr(self.cassandra_meta, 'clustering_keys'):
            self.clustering_keys = (
                self.query.model.CassandraMeta.clustering_keys
            )

        else:
            self.clustering_keys = []

        self.cql_query = self.column_family_class.objects.values_list(
            *self.column_names
        )

    @property
    def filterable_columns(self):
        return itertools.chain(
            self.clustering_keys,
            self.indexed_columns
        )

    def _get_rows_by_indexed_column(self, range_predicates):
        # Let's sort the predicates in efficient order.
        sorted_predicates = []
        predicates_by_column = {
            predicate.column: predicate for predicate in range_predicates
        }
        if self.pk_column in predicates_by_column:
            sorted_predicates.append(
                predicates_by_column[self.pk_column]
            )

        for column in self.clustering_keys:
            if column in predicates_by_column:
                sorted_predicates.append(
                    predicates_by_column[column]
                )

        assert len(range_predicates) == len(sorted_predicates)

        query = self.cql_query
        for predicate in sorted_predicates:
            if predicate._is_exact():
                query = query.filter(**{
                    predicate.column: predicate.start
                })

            else:
                if None is predicate.start:
                    start_op = None

                else:
                    if predicate.start_inclusive:
                        start_op = 'gte'
                    else:
                        start_op = 'gt'
                    start_op = '__'.join([
                        predicate.column,
                        start_op
                    ])

                if None is predicate.end:
                    end_op = None

                else:
                    if predicate.end_inclusive:
                        end_op = 'lte'
                    else:
                        end_op = 'lt'
                    end_op = '__'.join([
                        predicate.column,
                        end_op
                    ])

                if None is not start_op:
                    query = query.filter(**{
                        start_op: predicate.start
                    })

                if None is not end_op:
                    query = query.filter(**{
                        end_op: predicate.end
                    })
                
        return query
    
    def get_row_range(self, range_predicates):
        '''
        !!! Does this need to check for clustering key? !!!
        '''
        if not isinstance(range_predicates, list):
            range_predicates = list(range_predicates)

        for predicate in range_predicates:
            assert(
                predicate.column == self.pk_column or
                predicate.column in self.filterable_columns
            )

        return self._get_rows_by_indexed_column(range_predicates)
    
    def get_all_rows(self):
         return self.cql_query.all()
    
    def _get_query_results(self):
        if self.cache == None:
            assert(self.root_predicate != None)
            self.cache = self.root_predicate.get_matching_rows(self)

        return self.cache
    
    @safe_call
    def fetch(self, low_mark, high_mark):
        if self.root_predicate == None:
            raise DatabaseError('No root query node')
        
        if high_mark is not None and high_mark <= low_mark:
            return

        results = self._get_query_results()

        if low_mark is not None or high_mark is not None:
            results = results[low_mark:high_mark]

        

        for entity in results:
            yield entity

    def count(
        self,
        limit=None
    ):
        return self.cql_query.count()

    def delete(
        self,
        columns=set()
    ):
        return self.cql_query.delete()

    def order_by(
        self,
        ordering
    ):
        if isinstance(ordering, bool):
            self.reverse_order = not ordering
            return

        for order in ordering:
            if isinstance(order, basestring):
                if order.startswith('-'):
                    field_name = order[1:]
                    ascending = False

                else:
                    field_name = order
                    ascending = True

            elif 2 == len(order):
                field, ascending = order
                field_name = field.name

            else:
                raise ProgrammingError(
                    'Invalid ordering specification: %s' % order,
                )

            order_string = ''.join([
                '' if ascending else '-',
                field_name
            ])

            partition_key_filtered = (
                self.pk_column in [filter[0] for filter in self.filters]
            )

            if (
                not partition_key_filtered or
                field_name not in self.clustering_keys
            ):
                for order in self.ordering:
                    self.add_inefficient_order_by(order)

                self.add_inefficient_order_by(order_string)

            else:
                self.ordering.append(order_string)

    @safe_call
    def add_inefficient_order_by(self, ordering):
        if ordering.startswith('-'):
            field_name = ordering[1:]
            reversed = True
        else:
            field_name = ordering
            reversed = False

        self.inefficient_ordering.append((field_name, reversed))
            
    def init_predicate(self, parent_predicate, node):
        if isinstance(node, WhereNode):
            if node.connector == OR:
                compound_op = COMPOUND_OP_OR
            elif node.connector == AND:
                compound_op = COMPOUND_OP_AND
            else:
                raise InvalidQueryOpException()

            predicate = CompoundPredicate(compound_op, node.negated)

            for child in node.children:
                child_predicate = self.init_predicate(predicate, child)

            if parent_predicate:
                parent_predicate.add_child(predicate)
        else:
            decoded_child = self._decode_child(node)
            assert parent_predicate

            parent_predicate.add_filter(*decoded_child)
            self.filters.append(decoded_child)
            predicate = None
            
        return predicate
    
    # FIXME: This is bad. We're modifying the WhereNode object that's passed in to us
    # from the Django ORM. We should do the pruning as we build our predicates, not
    # munge the WhereNode.
    def remove_unnecessary_nodes(self, node, retain_root_node):
        if isinstance(node, WhereNode):
            child_count = len(node.children)
            for i in range(child_count):
                node.children[i] = self.remove_unnecessary_nodes(node.children[i], False)
            if (not retain_root_node) and (not node.negated) and (len(node.children) == 1):
                node = node.children[0]
        return node
        
    @safe_call
    def add_filters(self, filters):
        """
        Traverses the given Where tree and adds the filters to this query
        """

        assert isinstance(filters,WhereNode)
        self.remove_unnecessary_nodes(filters, True)
        self.root_predicate = self.init_predicate(None, filters)


class SQLCompiler(NonrelCompiler):
    query_class = CassandraQuery

    def execute_sql(
        self,
        result_type=MULTI
    ):
        return super(SQLCompiler, self).execute_sql(result_type)


class SQLInsertCompiler(NonrelInsertCompiler, SQLCompiler):
    def insert(
        self,
        values,
        return_id
    ):
        meta = self.query.get_meta()

        column_family = get_column_family(
            self.connection,
            self.query.model
        )

        inserted_row_keys = []

        with BatchQuery() as b:
            for row in values:
                # HELP!:Is this the right place for this logic? I can't seem
                # to figure out where Django adds the value for auto fields.
                if meta.has_auto_field and meta.pk.column not in row.keys():
                    row[meta.pk.column] = uuid4()

                inserted = column_family.batch(b).create(
                    **row
                )

                inserted_row_keys.append(inserted.pk)

        if return_id:
            if len(inserted_row_keys) == 1:
                return inserted_row_keys[0]

            else:
                return inserted_row_keys


class SQLUpdateCompiler(NonrelUpdateCompiler, SQLCompiler):
    def update(
        self,
        values
    ):
        return NonrelUpdateCompiler.update(self, values)


class SQLDeleteCompiler(NonrelDeleteCompiler, SQLCompiler):
    pass



from distutils.core import setup

setup(
    name='djangocassandra',
    version='0.0.2',
    description='Cassandra support for the Django web framework',
    long_description=(
        'The Cassandra database backend for Django has been '
        'developed to allow developers to utilize the Apache '
        'Cassandra Database in their Django projects.'
    ),
    author='Seth Denner',
    author_email='seth@knotis.com',
    maintainer='Knotis Inc.',
    maintainer_email='support@knotis.com',
    url='https://github.com/Knotis/djangocassandra',
    download_url='https://github.com/Knotis/djangocassandra/archive/master.zip',
    license='BSD License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django :: 1.7',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'Topic :: Internet',
        
    ],
    packages=[
        'djangocassandra',
        'djangocassandra.db',
        'djangocassandra.db.backends',
        'djangocassandra.db.backends.cassandra'
    ],
    install_requires=[
        'django>=1.7, < 1.8',
        'djangotoolbox>=1.6.2, < 1.7',
        'cassandra-driver>=2.1.4',
    ],
)



from cassandra.cqlengine import columns
from cassandra.cqlengine.models import (
    Model as CqlEngineModel,
    ModelMetaClass as CqlEngineModelMetaClass
)
from cassandra.cqlengine.management import create_keyspace


internal_type_to_column_map = {
    'AutoField': columns.UUID,
    'RelatedAutoField': columns.UUID,
    'ForeignKey': columns.UUID,
    'OneToOneField': columns.UUID,
    'ManyToManyField': columns.UUID,
    'BigIntegerField': columns.BigInt,
    'BooleanField': columns.Boolean,
    'CharField': columns.Text,
    'CommaSeparatedIntegerField': columns.Text,
    'DateField': columns.Date,
    'DateTimeField': columns.DateTime,
    'DecimalField': columns.Decimal,
    'EmailField': columns.Text,
    'FileField': columns.Text,
    'FilePathField': columns.Text,
    'FloatField': columns.Float,
    'ImageField': columns.Text,
    'IntegerField': columns.Integer,
    'IPAddressField': columns.Inet,
    'NullBooleanField': columns.Boolean,
    'PositiveIntegerField': columns.Integer,
    'PositiveSmallIntegerField': columns.Integer,
    'SlugField': columns.Text,
    'SmallIntegerField': columns.Integer,
    'TextField': columns.Text,
    'TimeField': columns.DateTime,
    'URLField': columns.Text,
    'AbstractIterableField': columns.List,
    'ListField': columns.List,
    'SetField': columns.Set,
    'DictField': columns.Map,
    'EmbeddedModelField': columns.Map,
    'RawField': columns.Blob,
    'BlobField': columns.Blob
}


class CqlColumnFamilyMetaClass(CqlEngineModelMetaClass):
    __column_families__ = {}

    def __new__(
        meta,
        name,
        bases,
        attrs
    ):
        if name in CqlColumnFamilyMetaClass.__column_families__:
            return CqlColumnFamilyMetaClass.__column_families__[name]

        model = attrs.get('__model__')
        if hasattr(model, 'CassandraMeta'):
            cassandra_options = model.CassandraMeta

        else:
            cassandra_options = None

        if None is not model:
            '''
            Create primary/clustering keys first. 
            '''
            primary_key_field = model._meta.pk
            column_type = internal_type_to_column_map[
                primary_key_field.get_internal_type()
            ]
            column = column_type(
                primary_key=True
            )
            attrs[
                primary_key_field.db_column if
                primary_key_field.db_column else
                primary_key_field.column
            ] = column

            clustering_keys = []
            if cassandra_options and hasattr(cassandra_options, 'clustering_keys'):
                clustering_keys = cassandra_options.clustering_keys

                for column_name in cassandra_options.clustering_keys:
                    field = model._meta.get_field(column_name)
                    if field.column == primary_key_field.column:
                        # Skip primary key if it was included in the clustering keys.
                        continue

                    column_type = internal_type_to_column_map[
                        field.get_internal_type()
                    ]
                    column = column_type(
                        primary_key=True
                    )
                    attrs[
                        field.db_column if
                        field.db_column else
                        field.column
                    ] = column
            
            for field in model._meta.fields:
                field_name = (
                    field.db_column if
                    field.db_column else
                    field.column
                )
                if field.primary_key or field_name in clustering_keys:
                    continue
                
                column_type = internal_type_to_column_map[
                    field.get_internal_type()
                ]
                column = column_type(
                    index=field.db_index,
                    db_field=field.db_column,
                    required=not field.blank
                )
                attrs[
                    field.db_column if
                    field.db_column else
                    field.column
                ] = column

        column_family = super(CqlColumnFamilyMetaClass, meta).__new__(
            meta,
            name,
            bases,
            attrs
        )

        CqlColumnFamilyMetaClass.__column_families__[name] = column_family
        return column_family


class CqlColumnFamily(CqlEngineModel):
    __model__ = None
    __abstract__ = True
    __metaclass__ = CqlColumnFamilyMetaClass


'''
Table Options:

'comment': None,
'read_repair_chance': None,
'dclocal_read_repair_chance': None,
'replicate_on_write': None,
'gc_grace_seconds': None,
'bloom_filter_fp_chance': None,
'caching': None,
'compaction_strategy_class': None,
'compaction_strategy_options': None,
'min_compaction_threshold': None,
'max_compaction_threshold': None,
'compression_parameters': None,
'min_index_interval': None,
'max_index_interval': None,
'index_interval': None,
'speculative_retry': None,
'rows_per_partition_to_cache': None,
'memtable_flush_period_in_ms': None,
'populate_io_cache_on_flush': None,
'compaction': None,
'compression': None,
'default_time_to_live': None
'''
default_table_options = {
}


default_cassandra_model_settings = {
    'table_options': default_table_options
}


def get_column_family(
    connection,
    model
):
    connection_settings = connection.settings_dict

    if hasattr(model, 'Cassandra'):
        cassandra_settings = model.Cassandra.__dict__

    else:
        cassandra_settings = default_cassandra_model_settings

    if hasattr(cassandra_settings, 'keyspace'):
        keyspace = cassandra_settings.keyspace

    else:
        keyspace = connection_settings.get('DEFAULT_KEYSPACE')

    keyspace_settings = connection_settings.get('KEYSPACES', {}).get(keyspace)
    if None is keyspace_settings:
        keyspace_settings = {}  # Replace with default keyspace settings.

    replication_factor = keyspace_settings.get(
        'replication_factor',

    )

    replication_strategy_class = keyspace_settings['strategy_class']

    create_keyspace(
        keyspace,
        replication_strategy_class,
        replication_factor
    )

    table_options = default_table_options

    if hasattr(cassandra_settings, 'table_options'):
        if not isinstance(cassandra_settings.table_optoins, dict):
            raise DatabaseError(
                'The value of table_optoins in the Cassandra class '
                'must be a dict containing overrides for the default'
                'column family options.'
            )
        table_options.update(cassandra_settings.table_metadata)

    return type(
        str(model._meta.db_table),
        (CqlColumnFamily,), {
            '__model__': model,
            'table_options': table_options
        }
    )



from django.db.models.sql.where import (
    WhereNode,
    EverythingNode
)


class CQLWhereNode(WhereNode):
    def as_cql(
        self,
        qn,
        connection
    ):
        return self.as_sql(
            qn,
            connection
        )


class CQLEverythingNode(EverythingNode):
    pass
